/*
 * JBoss, Home of Professional Open Source
 * Copyright 2013, Red Hat, Inc. and/or its affiliates, and individual
 * contributors by the @authors tag. See the copyright.txt in the
 * distribution for a full listing of individual contributors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jboss.as.quickstarts.ejb.remote.client;

import org.jboss.as.quickstarts.ejb.remote.stateful.RemoteCounter;
import org.jboss.as.quickstarts.ejb.remote.stateless.RemoteCalculator;
import org.jboss.security.ClientLoginModule;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.AppConfigurationEntry;
import javax.security.auth.login.Configuration;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import javax.security.sasl.RealmCallback;

import java.io.IOException;
import java.security.AccessController;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

/**
 * A sample program which acts a remote client for a EJB deployed on AS7 server. This program shows how to lookup stateful and
 * stateless beans via JNDI and then invoke on them
 * 
 * @author Jaikiran Pai
 */
public class RemoteEJBClient {

	private static final String username = "dumbledore";
	private static final String password = "harrypotter123";
	private static String realm = "other";
	
	private static void setupJndiProperties(Hashtable<String,Object> props)
	{
        props.put(Context.URL_PKG_PREFIXES, "org.jboss.ejb.client.naming");
        
        // Is this really necessary and shouldn't ClientLoginModule in conjunction with LoginContext be enough?
        // still doesn't work when I put it in.
//        props.put("remote.connection.username", username);
//        props.put("remote.connection.password", password);
//        props.put("java.naming.security.principal", username);
//        props.put("java.naming.security.credentials", password);
	}
	
	private static boolean secureClient()
	{
	        System.out.println("Preparing LoginContext.");

	        CallbackHandler cbh = new CallbackHandler() {
	            public void handle(Callback[] callbacks) throws IOException, UnsupportedCallbackException {
	                for (Callback current : callbacks) {
	                	if (current instanceof RealmCallback) {
                            ((RealmCallback) current).setText(realm);
	                	} else if (current instanceof NameCallback) {
	                        ((NameCallback) current).setName(username);
	                    } else if (current instanceof PasswordCallback) {
	                        ((PasswordCallback) current).setPassword(password.toCharArray());
	                    } else {
	                        throw new UnsupportedCallbackException(current);
	                    }
	                }
	            }
	        };
	        Configuration config = new Configuration() {

	            @Override
	            public AppConfigurationEntry[] getAppConfigurationEntry(String name) {
	                if (realm.equals(name) == false) {
	                    throw new IllegalArgumentException("Unexpected configuration name '" + name + "'");
	                }
	                Map<String, String> options = new HashMap<String, String>();
// I'm not multi-threaded so I shouldn't have to put in these options.	                
//	                options.put("multi-threaded", "true");
//	                options.put("restore-login-identity", "true");

	                AppConfigurationEntry clmEntry = new AppConfigurationEntry(ClientLoginModule.class.getName(),
	                    AppConfigurationEntry.LoginModuleControlFlag.REQUIRED, options);

	                return new AppConfigurationEntry[] { clmEntry };
	            }
	        };

	        LoginContext context;
			try {
				context = new LoginContext(realm, new Subject(), cbh, config);
				System.out.println("logging in.");
				context.login();
			} catch (LoginException e) {
				System.out.println("Failed to establish login context: " + e.getMessage());
				return false;
			}
			
			
			
	        Subject subject = context.getSubject();
	        if (subject != null)
	        	System.out.println("LoginContext Subject" + subject.toString());
	        
	        Subject current = Subject.getSubject(AccessController.getContext());  // get's the current subject...
	        if (current != null)
	        	System.out.println("CurrentSubject: " + current.toString());
	        
// I've gotten this before and checked and it was correct...do I need to do more to setup an initial context for propagation?	        
//	        Subject c2 = SecurityContextAssociation.getSubject();
//	        if (c2 != null)
//	        	System.out.println("CurrentSubject: " + c2.toString());
	        
	        
	        return true;
		
	}
	

    public static void main(String[] args) throws Exception {
    	
    	if (secureClient())
    	{
	        // Invoke a stateless bean
	        invokeStatelessBean();
	
	        // Invoke a stateful bean
	        invokeStatefulBean();
    	}
    }

    /**
     * Looks up a stateless bean and invokes on it
     * 
     * @throws NamingException
     */
    private static void invokeStatelessBean() throws NamingException {
        // Let's lookup the remote stateless calculator
        final RemoteCalculator statelessRemoteCalculator = lookupRemoteStatelessCalculator();
        System.out.println("Obtained a remote stateless calculator for invocation");
        // invoke on the remote calculator
        int a = 204;
        int b = 340;
        System.out.println("Adding " + a + " and " + b + " via the remote stateless calculator deployed on the server");
        int sum = statelessRemoteCalculator.add(a, b);
        System.out.println("Remote calculator returned sum = " + sum);
        if (sum != a + b) {
            throw new RuntimeException("Remote stateless calculator returned an incorrect sum " + sum + " ,expected sum was "
                    + (a + b));
        }
        // try one more invocation, this time for subtraction
        int num1 = 3434;
        int num2 = 2332;
        System.out.println("Subtracting " + num2 + " from " + num1
                + " via the remote stateless calculator deployed on the server");
        int difference = statelessRemoteCalculator.subtract(num1, num2);
        System.out.println("Remote calculator returned difference = " + difference);
        if (difference != num1 - num2) {
            throw new RuntimeException("Remote stateless calculator returned an incorrect difference " + difference
                    + " ,expected difference was " + (num1 - num2));
        }
    }

    /**
     * Looks up a stateful bean and invokes on it
     * 
     * @throws NamingException
     */
    private static void invokeStatefulBean() throws NamingException {
        // Let's lookup the remote stateful counter
        final RemoteCounter statefulRemoteCounter = lookupRemoteStatefulCounter();
        System.out.println("Obtained a remote stateful counter for invocation");
        // invoke on the remote counter bean
        final int NUM_TIMES = 5;
        System.out.println("Counter will now be incremented " + NUM_TIMES + " times");
        for (int i = 0; i < NUM_TIMES; i++) {
            System.out.println("Incrementing counter");
            statefulRemoteCounter.increment();
            System.out.println("Count after increment is " + statefulRemoteCounter.getCount());
        }
        // now decrementing
        System.out.println("Counter will now be decremented " + NUM_TIMES + " times");
        for (int i = NUM_TIMES; i > 0; i--) {
            System.out.println("Decrementing counter");
            statefulRemoteCounter.decrement();
            System.out.println("Count after decrement is " + statefulRemoteCounter.getCount());
        }
    }

    /**
     * Looks up and returns the proxy to remote stateless calculator bean
     * 
     * @return
     * @throws NamingException
     */
    private static RemoteCalculator lookupRemoteStatelessCalculator() throws NamingException {
        final Hashtable<String,Object> jndiProperties = new Hashtable<String,Object>();
        setupJndiProperties(jndiProperties);
        final Context context = new InitialContext(jndiProperties);

        // The JNDI lookup name for a stateless session bean has the syntax of:
        // ejb:<appName>/<moduleName>/<distinctName>/<beanName>!<viewClassName>
        //
        // <appName> The application name is the name of the EAR that the EJB is deployed in
        // (without the .ear). If the EJB JAR is not deployed in an EAR then this is
        // blank. The app name can also be specified in the EAR's application.xml
        //
        // <moduleName> By the default the module name is the name of the EJB JAR file (without the
        // .jar suffix). The module name might be overridden in the ejb-jar.xml
        //
        // <distinctName> : AS7 allows each deployment to have an (optional) distinct name.
        // This example does not use this so leave it blank.
        //
        // <beanName> : The name of the session been to be invoked.
        //
        // <viewClassName>: The fully qualified classname of the remote interface. Must include
        // the whole package name.

        // let's do the lookup
        return (RemoteCalculator) context.lookup("ejb:/jboss-as-ejb-remote-server-side/CalculatorBean!"
                + RemoteCalculator.class.getName());
    }

    /**
     * Looks up and returns the proxy to remote stateful counter bean
     * 
     * @return
     * @throws NamingException
     */
    private static RemoteCounter lookupRemoteStatefulCounter() throws NamingException {
        final Hashtable<String,Object> jndiProperties = new Hashtable<String,Object>();
        setupJndiProperties(jndiProperties);
        final Context context = new InitialContext(jndiProperties);

        // The JNDI lookup name for a stateful session bean has the syntax of:
        // ejb:<appName>/<moduleName>/<distinctName>/<beanName>!<viewClassName>?stateful
        //
        // <appName> The application name is the name of the EAR that the EJB is deployed in
        // (without the .ear). If the EJB JAR is not deployed in an EAR then this is
        // blank. The app name can also be specified in the EAR's application.xml
        //
        // <moduleName> By the default the module name is the name of the EJB JAR file (without the
        // .jar suffix). The module name might be overridden in the ejb-jar.xml
        //
        // <distinctName> : AS7 allows each deployment to have an (optional) distinct name.
        // This example does not use this so leave it blank.
        //
        // <beanName> : The name of the session been to be invoked.
        //
        // <viewClassName>: The fully qualified classname of the remote interface. Must include
        // the whole package name.

        // let's do the lookup
        return (RemoteCounter) context.lookup("ejb:/jboss-as-ejb-remote-server-side/CounterBean!"
                + RemoteCounter.class.getName() + "?stateful");
    }
}
